ReturnTypes = ['text', 'src', 'attr', 'element']

ActionTypes = [
    "ClickAction",
    "InputAction",
    "CaptureAction",
    "PublishAction"
]


def get_mandatory_params(actionType):

    paramsMap = dict(Action=["css","xpath","text","isSingle", "actionType"],
                     CaptureAction=['captureName'],
                     ClickAction=[],
                     InputAction=['inputString'],
                     PublishAction=['urlStub'])
    return list(set(paramsMap.get('Action') + paramsMap.get(actionType, [])))





import logging
import os
import traceback
import requests
import json
from json.encoder import JSONEncoder
from feed.actiontypes import ReturnTypes
from bs4 import BeautifulSoup, Tag
from feed.settings import kafka_params, routing_params
import signal


from kafka import KafkaConsumer, KafkaProducer



class GracefulKiller:
    kill_now = False
    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self,signum, frame):
        self.kill_now = True



class ObjectSearchParams:
    def __init__(self, **kwargs):
        #super().__init__(**kwargs)
        self.isSingle = kwargs.get('isSingle', False)
        self.returnType = kwargs.get('returnType', 'src')
        self.attribute = kwargs.get('attribute', None)

    def __dict__(self):
        return self.kwargs

    def _verifyResultLength(self, items):
        if len(items) == 0:
            return False
        if self.isSingle and len(items) > 1:
            self.backup = items
            return False
        else:
            return True

    def _returnItem(item, driver):
        raise NotImplementedError
    def search(item, driver):
        raise NotImplementedError


class BrowserSearchParams(ObjectSearchParams):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.kwargs = kwargs
        self.css = kwargs.get('css')
        self.xpath = kwargs.get('xpath')
        self.text = kwargs.get('text')
        #self.text = kwargs.get('class')
        self.backup = None

    def _returnItem(self, item, driver):
        logging.debug(f'returning returnType=[{self.returnType}] type for actionType=[{type(self).__name__}]')
        if self.returnType == 'text':
            formatted = lambda item: item.text
        elif self.returnType == 'src':
            classes = set([element.get_attribute('class') for element in item])
            soup = BeautifulSoup(driver.page_source)
            out = []
            for cls in classes:
                logging.debug(f'ObjectSearchParam::_returnItem(): searching for node with attribues, class=[{cls}]')
                out.extend(soup.findAll(attrs={'class': cls}))
                item = out
            formatted = lambda item: item
        elif self.returnType == 'attr':
            formatted = lambda element: element.get_attribute(self.attribute)
        elif self.returnType == 'element':
            formatted = lambda item: item
        logging.debug(f'ObjectSearchParams::_returnItem(): returning {len(item)}.')
        return list(map(formatted, item)) if len (item) > 1 else formatted(item[0])

    def search(self, driver):

        # first try css selector
        logging.info(f'{type(self).__name__}::search(driver): searching for elemnent with css=[{self.css}]')
        ret = driver.find_elements_by_css_selector(self.css)
        if self._verifyResultLength(ret):
            logging.info(f'BrowserSearchParams::search(): found elements count=[{len(ret)}], isSingle=[{self.isSingle}] with css')
            return self._returnItem(ret, driver)

        # then try xpath
        logging.info(f'{type(self).__name__}::search(driver): searching for elemnent with xpath=[{self.xpath}]')
        ret = driver.find_elements_by_xpath(self.xpath)
        if self._verifyResultLength(ret):
            logging.info(f'BrowserSearchParams::search: found element [{ret}] with xpath')
            return self._returnItem(ret, driver)

        # TODO: should search backup list with text at this point
        if self.backup:
            for res in filter(lambda item: item.text.upper() == self.text, self.backup):
                logging.debug(f'using element [{ret}] from backup')
                return self._returnItem([ret], driver)
            return self._returnItem([self.backup[0]], driver)
        else:
            # TODO brute search with text at this point
            return None


class Action(BrowserSearchParams):

    def __init__(self, position, **kwargs):
        super().__init__(**kwargs)
        self.kwargs = kwargs
        self.position = position

    @staticmethod
    def execute(chain, action):
        actionType = type(action).__name__
        try:
            ret = getattr(chain, f'on{actionType}')(action)
            logging.debug(f'Action::execute: Action executed succesfully, name=[{chain.name}], position=[{action.position}]')
            return ret
        except Exception as ex:
            traceback.print_exc()
            logging.warning(f'Action::execute:: {type(ex).__name__} thrown whilst processing name=[{chain.name}], position=[{action.position}], args=[{ex.args}]')
            return False
            # TODO Exception reporting callback called here
            # OnClickException for example

    def getActionableItem(self, action, driver):
        item = self.search(driver)
        logging.info(f'{type(self).__name__}::getActionableItem: have num_items=[{ 1 if not isinstance(item, list) else len(item)}]')
        return item

    @staticmethod
    def get_params():
        # TODO For UI-Server
        return [self.__dict__().keys()]


class CaptureAction(Action):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.kwargs.pop('data', None)
        self.kwargs.pop('src', None)
        self.returnType = 'src'
        self.captureName = kwargs['captureName'] # mandatory
        self.data = kwargs.get('data', None)

    def __dict__(self):
        return dict(data=self.data, **self.kwargs)

class InputAction(Action):

    def __init__(self, inputString, **kwargs):
        super().__init__(**kwargs)
        self.insputString = inputString
        self.isSingle = True
        self.returnType = 'element'

    def __dict__(self):
        return dict(inputString=self.insputString, **self.kwargs)

class PublishAction(Action):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.urlStub = kwargs.get('urlStub')

    def __dict__(self):
        return dict(url=self.url, urlStub=self.urlStub, **self.kwargs)

class ClickAction(Action):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.isSingle = True
        self.returnType = 'element'

ActionTypes = {
    "ClickAction": ClickAction,
    "InputAction": InputAction,
    "CaptureAction": CaptureAction,
    "PublishAction": PublishAction
}

class ActionChain:
    actions= {}

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.name = kwargs.get('name')
        self.startUrl = kwargs.get('startUrl')
        self.repeating = kwargs.get('isRepeating', True)
        actionParams = kwargs.get('actions', [])
        self.failedChain = False
        for order, params in enumerate(actionParams):
            try:
                action = ActionChain.actionFactory(position=order, actionParams=params)
                self.actions.update({order: action})
            except KeyError as ex:
                # TODO: wAt this point we should pass this onto the user
                traceback.print_exc()
                logging.error(f'{type(self).__name__}::__init__(): chainName=[{self.name}], position=[{order}] actionType=[{params.get("actionType")}] is missing {ex.args} default parameter')


    def recoverHistory(self):
        try:
            req = requests.get('http://{host}:{port}/routingcontroller/getLastPage/{name}'.format(name=self.name, **routing_params))
        except Exception as e:
            logging.warning(f'ActionChain::recoverHistory: router is unavailable')
        logging.info(f'{type(self).__name__}::recoverHistory have {req} from routing.')
        try:
            data = req.json()
        except Exception:
            logging.warning(f'Did not get valid response from router. response=[{req}]')
        return data

    def saveHistory(self, url):
        pass

    @staticmethod
    def actionFactory(position, actionParams):
        logging.info(f'ActionChain::actionFactory: node=[{position}]: {", ".join(map(lambda key: "{}=[{}]".format(key, actionParams[key]), actionParams))}')
        actionConstructor = ActionTypes.get(actionParams.get('actionType'))
        return actionConstructor(position=position, **actionParams)

    def initialise(self, caller):
        pass

    """
    following methods correspond to module://feed.actiontypes.ActionTypes
    """
    def onClickAction(self, action):
        raise NotImplementedError
    def onInputAction(self, action):
        raise NotImplementedError
    def onPublishAction(self, action):
        raise NotImplementedError
    def onCaptureAction(self, action):
        raise NotImplementedError

    def onChainEnd(self):
        """
        called after the chaions execute method
        """
        logging.info(f'{type(self).__name__}::onChainEnd(): chainName={self.name}')

    def execute(self, caller, initialise=True):
        self.failedChain = False
        if initialise:
            self.initialise(caller)
        for i in range(len(self.actions)):
            self.current_pos = i
            action = self.actions.get(i)
            logging.info(f'ActionChain::execute(): executing action {type(action).__name__}')
            success = Action.execute(self, action)
            if not success:
                logging.info(f'{type(self).__name__}::execute(): Detected failure: actionType={type(action).__name__}, position={i}, name={self.name}. Will go straight to next action. {"Will not re-evaluate" if self.repeating else ""}')
                self.failedChain = True
                continue
            callBackMethod = getattr(caller, f'on{type(action).__name__}Callback')
            for item in success:
                callBackMethod(item)
            self.saveHistory()
            self.onChainEnd()

    def getRepublishRoute(self, action):
        name = 'unknown'
        if isinstance(action, CaptureAction):
            name = 'summarizer-route'
        if isinstance(action, PublishAction):
            name = 'worker-route'
        if isinstance(action, ActionChain):
            # TODO: If publisher wants to invoke a chain to be ran, they should set action to be a new action chain object
            name = 'leader-route'
        return f'{os.getenv("KAFKA_TOPIC_PREFIX", "d")}-{name}'

    def rePublish(self, action, *args, **kwargs):
        pass


class KafkaChainPublisher(ActionChain):
    pass


class ActionReturnSerialiser(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Tag):
            return str(obj)
        elif isinstance(obj, Action):
            return obj.__dict__()
        else:
            return super().default(obj)


class KafkaActionPublisher(ActionChain):

    flush_rate = os.getenv('PRODUCER_FLUSH_RATE', 30)
    def __init__(self):
        self.producer = KafkaProducer(**kafka_params, value_serializer=lambda m: bytes(json.dumps(m, cls=ActionReturnSerialiser).encode('utf-8')))
        self.messages_out_since_flush = 0

    def rePublish(self, actionReturn):
        self.messages_out_since_flush += 1
        topic = self.getRepublishRoute(actionReturn.action)
        # construct chain parameters to send
        actionReturn.action.data = actionReturn.data
        payload = {
            'actions': [actionReturn.action],
            'startUrl': actionReturn.current_url,
            'isRepeating': False,
            'name': actionReturn.name
        }

        self.producer.send(topic=topic, value=payload)
        logging.info(f'{type(self).__name__}::rePublish(): Republished ActionReturn(startUrl={actionReturn.current_url} for {type(actionReturn.action).__name__} to {topic}')
        # check to see if we should flush
        if self.messages_out_since_flush >= self.flush_rate:
            self.messages_out_since_flush = 0
            logging.info('KafkaActionSubscription::rePublish: flushing messages')
            self.producer.flush()


class ActionChainRunner:

    driver = None
    def __init__(self, implementation, **kwargs):
        logging.info(f'initiated ActionChainRunner fo {type(implementation).__name__}')
        self.implementation = implementation

    def subscription(self):
        pass

    def onClickActionCallback(self, *args, **kwargs):
        logging.info(f'onClickActionCallback')

    def onInputActionCallback(self, *args, **kwargs):
        logging.info(f'onInputActionCallback')

    def onPublishActionCallback(self, *args, **kwargs):
        logging.info(f'onPublishActionCallback')

    def onCaptureActionCallback(self, *args, **kwargs):
        logging.info(f'onCaptureActionCallback')

    def onChainEndCallback(self, chain, chainReturn):
        logging.info(f'onChainEndCallback')

    def renewDriverSession(self):
        pass

    def driverHealthCheck(self):
        pass

    def initialiseCallback(self, *args, **kwargs):
        logging.info('initialiseCallback')

    def main(self):
        killer = GracefulKiller()
        logging.info(f'{type(self).__name__}::main(): beginning subscription poll of kafka')
        for actionChainParams in self.subscription():
            if not self.driverHealthCheck():
                self.renewDriverSession()
            if killer.kill_now:
                self.cleanUp()
                logging.info(f'cleaned up resources')
                break
            actionChain = self.implementation(driver=self.driver, **actionChainParams)
            logging.info(f'{type(self).__name__}::main(): START:{actionChain.name} implementing action chain {actionChainParams.get("name")}: {json.dumps(actionChainParams, indent=4)}')
            ret = actionChain.execute(self)
            self.onChainEndCallback(actionChain, ret)
            # should the chain be automatically be re ran from where we are?
            # this can be disabled in the implementation of onChainEndCallback
            while actionChain.repeating and not actionChain.failedChain:
                actionChain.execute(caller=self, initialise=False)
            logging.info(f'{type(self).__name__}::main(): END:{actionChain.name} ActionChain::execute() has returned')

    def cleanUp(self):
        logging.warning(f'ActionChainRunner::cleanUp() No cleanup has been implemented')


class KafkaActionSubscription(ActionChainRunner):

    def __init__(self, topic,  **kwargs):
        super().__init__(**kwargs)
        logging.info(f'Starting ActionChainRuner type {type(self).__name__}, topic=[{topic}]')
        self._consumer = KafkaConsumer(**kafka_params, value_deserializer=lambda m: json.loads(m.decode('utf-8')))
        self._consumer.subscribe([f'{os.getenv("KAFKA_TOPIC_PREFIX")}-{topic}'])

    def subscription(self):
        for mes in self._consumer:
            yield mes.value


class CommandsActionSubscription(ActionChainRunner):

    def __init__(self, endpoint, actionsImpl):
        super().__init__(actionsImpl)
        self.endpoint = endpoint

    def subscription(self):
        while True:
            actionReq = requests.get(endpoint)
            action = actionReq.json()
            yield action


ActionTypesMap = {
    "ClickAction": ClickAction,
    "InputAction": InputAction,
    "CaptureAction": CaptureAction,
    "PublishAction": PublishAction
}



import logging
import sys
import os

logger = logging.getLogger(__name__)
ch = logging.StreamHandler(stream=sys.stderr)
ch.setLevel(os.getenv("LOG_LEVEL", "INFO"))

logger.setLevel(os.getenv("LOG_LEVEL", "INFO"))

def getLogger(name, toFile=False):
    logger = logging.getLogger(name)
    feedLogger = logging.getLogger('feed')
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s -%(message)s |%(filename)s:%(lineno)d')
    ch.setLevel(os.getenv("LOG_LEVEL", "DEBUG"))
    ch.setFormatter(formatter)
    logger.setLevel(os.getenv("LOG_LEVEL", "INFO"))

    # logger.addHandler(ch)
    if not os.path.exists('/tmp/logs/'):
        os.mkdir('/tmp/logs')
    if toFile:
        fileHandler = logging.FileHandler(f'/tmp/logs/{name}.log')
        fileHandler.setFormatter(formatter)
        fileHandler.setLevel(os.getenv('FILE_LOGLEVEL', 'INFO'))
        logger.addHandler(fileHandler)
    feedLogger.addHandler(ch)
    return logger

def initialiseSrcLogger():
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setLevel(os.getenv("LOG_LEVEL", "DEBUG"))
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s -%(message)s |%(filename)s:%(lineno)d')
    handler.setFormatter(formatter)
    srcLogger = logging.getLogger('src.main')
    srcLogger.addHandler(handler)





from unittest import TestCase
import logging
import time
import unittest
import os
import docker
from feed.crawling import BrowserService
from feed.testinterfaces import SeleniumTestInterface


class TestBrowserService(SeleniumTestInterface, TestCase):

    def test_driverHealtCheck(self):
        res = self.browserService.driverHealthCheck()
        self.assertTrue(res)
        self.browserService.driver.close()
        res = self.browserService.driverHealthCheck()
        self.assertFalse(res)

    def testRenewWebCrawler(self):
        self.browserService.driver.close()
        self.browserService.renewWebCrawler()
        self.assertTrue(self.browserService.driverHealthCheck())

    def setUp(cls):
        cls.browserService = BrowserService()

    def tearDown(cls):
        cls.browserService.driver.quit()
        del cls.browserService




import os
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError
from pymongo.database import Database
from pymongo.collection import Collection
from datetime import timedelta
from flask.sessions import SessionInterface, SessionMixin
from flask import Flask, session
import time
import logging
from feed.settings import mongo_params


class ChainSession(SessionInterface):

    _sessioncollection: Collection = None
    _chaindefinitions: Collection = None
    sessionConstructor = None
    _client: MongoClient = None

    def __init__(self, sessionType):
        logging.info(f'Starting ChainSession mananager with sessionType=[{sessionType.__name__}] and parameters=[{mongo_params}]')
        self._client = MongoClient(**mongo_params)
        self.sessionConstructor = sessionType
        self._sessioncollection = self._client[os.getenv('CHAIN_DB', 'actionChains')][f'{sessionType.__name__}Sessions']
        self._chaindefinitions = self._client[os.getenv('CHAIN_DB', 'actionChains')]['actionChainDefinitions']

    def is_null_session(self, sessionObj):
        if sessionObj.name:
            logging.info(f'found null session {sessionObj.name}')
            return False
        else:
            return True

    def open_session(self, app, request):
        chainNames = request.path.split("/")
        name = self._chaindefinitions.find_one({'name': {"$in": chainNames}}, projection=["name"])
        if name is None:
            chainSession = self.sessionConstructor(name=None)
        else:
            logging.info(f'starting session for name=[{name.get("name")}]')
            chainSession = self._open_session(name=name.get('name'))

        chainSession.update({'chain_db': self._client[os.getenv('CHAIN_DB', 'sessions')],
                             'chainDefinitions': self._chaindefinitions})
        return chainSession

    def save_session(self, app, request, response):
        if session.modified:
            self._save_session(session)

    @staticmethod
    def get_session_id(chainName):
        return "{}-{}".format(chainName, time.strftime("%d_%m"))

    def _save_session(self, session):
        sessionOut = session.__dict__()
        sessionOut.update({'session_id': ChainSession.get_session_id(session.name)})
        logging.info(f'saving session for name=[{session.name}]')
        self._sessioncollection.replace_one({'session_id': ChainSession.get_session_id(session.name)}, replacement=sessionOut, upsert=True)

    def _open_session(self, name):
        # TODO: session id logic
        sess = self._sessioncollection.find_one({'session_id': ChainSession.get_session_id(name)})
        if sess is None:
            return self.sessionConstructor(name=name, session_id=ChainSession.get_session_id(name))
        else:
            return self.sessionConstructor(**sess)




#TODO def is_null_session():

def probeMongo(client):
    try:
        cl = client.server_info()
    except ServerSelectionTimeoutError as ex:
        logging.info(f'trying to connect to mongo with {mongo_params}')
        return False
    return True

def init_app(domainImpl):

    app = Flask(__name__)

    app.permanent_session_lifetime = timedelta(days=31)
    app.secret_key = os.getenv('SECRET_KEY', 'this is supposed to be secret')

    sessionManager = ChainSession(domainImpl)

    app.session_interface = sessionManager

    return app






from feed.logger import getLogger
from queue import Queue
import os
import sys
import traceback
from http.client import RemoteDisconnected
from time import sleep, time
import re
import requests as r
import selenium.webdriver as webdriver
from bs4 import BeautifulSoup
from bs4 import Tag, NavigableString
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from urllib3.exceptions import MaxRetryError, ProtocolError
import requests
import threading
import subprocess
import argparse

from feed.settings import nanny_params, browser_params, routing_params
from feed.service import Client

from feed.actionchains import ObjectSearchParams, ActionChain, ClickAction, InputAction, CaptureAction, PublishAction, Action

logging = getLogger(__name__)


def verifyUrl(url):
    if url is None:
        return False
    if ' ' in url:
        return False
    if 'http://' in url or 'https://' in url:
        return True
    if 'www' in url:
        return True


class BrowserActions(ActionChain):


    class Return:
        def __init__(self, action: Action, data, current_url, name, *args, **kwargs):
            self.name = name
            self.current_url = current_url
            self.data = data
            self.action = action

        def __dict__(self):
            return dict(name=self.name, current_url=self.current_url, data=self.data, action=self.action.__dict__())

    driver = None # type: WebDriver

    def __init__(self, driver: WebDriver, *args, **kwargs):
        super().__init__(*args, **kwargs)
        requests.get('http://{host}:{port}/routingcontroller/initialiseRoutingSession/{name}'.format(name=self.name, **routing_params))
        self.kwargs = kwargs
        self.driver = driver

    @staticmethod
    def _get_button_to_click(item, action):
        if not isinstance(item, Tag):
            return item
        if item.text and item.text.upper() == action.text.upper():
            return item
        logging.debug(f'checking Tag with text=[{item.text}]')
        for ch in item.children:
            logging.debug(f'##### {type(ch).__name__} #######')
            if isinstance(ch, NavigableString):
                ch = BrowserActions._searchNavigableStringForTag(ch, action.text)
                logging.debug(f'have navigable string')
                if ch is None:
                    continue
            if isinstance(ch, Tag):
                logging.debug(f'checking child {ch.text}')
            but = BrowserActions._get_button_to_click(ch, action)
            if but is not None:
                return but
        else:
            return None

    @staticmethod
    def _searchNavigableStringForTag(navString: NavigableString, text):
        if navString == text:
            return BeautifulSoup(str(navString))
        newString = navString.findNextSibling()
        if newString is not None:
            BrowserActions._searchNavigableStringForTag(newString, text)
        else:
            return None


    def _verify_class_string(self, item):

        if ' ' in item:
            return False
        else:
            return True

    def onClickAction(self, action: ClickAction):
        logging.info(f'{type(self).__name__}::onClickAction: css=[{action.css}], xpath=[{action.xpath}], text=[{action.text}]')
        button: WebElement = action.getActionableItem(action, self.driver)
        html_class = button.get_attribute('class')
        # TODO an action chain should cash there last interaction in the case of a repeat so it does not havbe to do the searching again.
        if not self._verify_class_string(html_class):
            logging.info(f'{type(self).__name__}::onClickAction: Checking {html_class} for {action.css}')
            html_class = list(filter(lambda item: item in action.css, html_class.split(' ')))[0]
        soup = BeautifulSoup(self.driver.page_source)
        logging.info(f'will search html with html_class=[{html_class}]')
        items = soup.find_all(attrs={'class': html_class})
        if len(items) < 1:
            logging.info(f'{type(self).__name__}::onClickAction: Couldnt find button with class=[{html_class}]')
            pass
        else:
            logging.info(f'{type(self).__name__}::onClickAction: items found from html parse. count={len(items)}')
            for item in items:
                altButton = self._get_button_to_click(item, action)
                if altButton and isinstance(altButton, Tag):
                    classNames = altButton.attrs.get('class')
                    logging.debug(f'checking class names, classNames=[{classNames}], len=[{len(classNames)}]')
                    for className in classNames:
                        elems = self.driver.find_elements_by_class_name(className)
                        logging.info(f'found {len(elems)} with className=[{className}]')
                        if len(elems) == 1:
                            logging.info(f'{type(self).__name__}: found unique button to click with className=[{className}], should only have appeared here once')
                            button = elems[0]
        logging.info(f'{type(self).__name__}::onClickAction(): clicking on text={button.text}')
        clickingFrom = self.driver.current_url
        clickTime = time()
        buttonTxt = button.text
        button.click()
        while ( self.driver.current_url == clickingFrom ) and (time() - clickTime <= 5):
            sleep(0.5)
            if self.driver.current_url == clickingFrom:
                logging.debug(f'{type(self).__name__}::onClickAction(): current url has not changed from clicking on "{buttonTxt}"')
            else:
                logging.debug(f'{type(self).__name__}::onClickAction(): current url has changed from "{clickingFrom}" to "{self.driver.current_url}"')
            logging.info(f'{type(self).__name__}::onClickAction(): current url has changed from "{clickingFrom}" to "{self.driver.current_url}"')
        return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action, data=None)]

    def onCaptureAction(self, action: CaptureAction):
        data = action.getActionableItem(action, self.driver)
        self.rePublish(key=self.driver.current_url, action=action, data=data)
        if not action.isSingle:
            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=item) for item in data]
        else:
            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=data)]

    def onPublishAction(self, action: PublishAction):
        logging.info(f'{type(self).__name__}::onPublishAction: css=[{action.css}], xpath=[{action.xpath}], text=[{action.text}]')
        data = action.getActionableItem(action, self.driver)
        logging.info(f'{type(self).__name__}::onPublishAction: have found data=[{len(data)}]')
        cls = data[0].get_attribute('class')
        soup = BeautifulSoup(self.driver.page_source)
        items = soup.findAll(attrs={'class': cls})
        out=[]
        for item in items:
            # try find a link on the top most surface of html
            link = item.attrs.get('href')
            if link and action.urlStub in link:
                # if it is in this add and continue
                out.append(link)
                continue
            logging.info(f'{type(self).__name__}::onPublishAction: Link was not immediately available, will try parent')
            # if failure, then try find a parent with regex
            parent = item.findParent(attrs={'href': re.compile(f'{action.urlStub}/*')})
            if parent is None:
                # if nothing found, then try find in children with regex
                child = item.findChild(attrs={'href': re.compile(f'{action.urlStub}/*')})
                logging.info(f'{type(self).__name__}::onPublishAction: Link was not immediately available, will try child')
            else:
                # if parent was found then add the link of it and continue
                logging.info(f'{type(self).__name__}::onPublishAction: Found parent link in parent tag.')
                out.append(parent.attrs.get('href'))
                continue
            if child is None: # were here because regex on parent was unsuccesful as well as the child
                # if link wasn't found in the child then try find 'a' tag of parent
                logging.info(f'{type(self).__name__}::onPublishAction: Link was not found with child, searching parent for a tag')
                parentAtag = item.findParent('a', attrs={'href': re.compile(f'{action.urlStub}/*')})
                link = ''
                if parentAtag:
                    # if found then take the link
                    logging.info(f'{type(self).__name__}::onPublishAction: Found parent "a" tag.')
                    link = parentAtag.attrs.get('href')
                if not parentAtag or action.urlStub not in link:
                    # if no suitable parent found with a tag or it was wrong, then try the same with the child
                    bckup = item.findChild('a', attrs={'href': re.compile(f'{action.urlStub}/*')})
                    logginginfo(f'{type(self).__name__}::onPublishAction: Checking backups for link.')
                    if bckup is not None:
                        link = bckup.attrs.get('href') if action.urlStub in bckup.attrs.get('href') else None
                # if all un succesful then dont add anything
                if link == '':
                    logging.warning(f'{type(self).__name__}::onPublishAction: could not find link to page item for chain=[{self.name}], action=[{action}]')
                    continue
                logging.info(f'{type(self).__name__}::onPublishAction: Found link=[{link}]. ')
                out.append(link)
            else:
                out.append(child.attrs.get('href'))
        # TODO should put this into a callback
        # TODO should republish chain here
        self.rePublish(key=self.driver.current_url, action=action, data=out)
        if not action.isSingle:
            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=url) for url in out]
        else:
            return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action,data=out[0])]

    def onInputAction(self, action: InputAction):
        inputField: WebElement = action.getActionableItem(action, self.driver)
        inputField.send_keys(action.inputString)
        logging.warning(f'{type(self).__name__}::onInputAction: chain=[{self.name}], action=[{action}]')
        return [BrowserActions.Return(current_url=self.driver.current_url, name=self.name, action=action, data=inputField)]

    def saveHistory(self):
        try:
            logging.info(f'BrowserActions::saveHistory: Saving current_url=[{self.driver.current_url}]')
            requests.get('http://{host}:{port}/routingcontroller/updateHistory/{name}'.format(name=self.name, **routing_params), data=self.driver.current_url)
        except Exception as e:
            logging.warning(f'BrowserActions::saveHistory: router is unavailable.')

    def initialise(self, caller):
        hist = self.recoverHistory()
        logging.info(f'BrowserActions::initialise: recovered history for {self.name}, url=[{hist}]')
        url = hist.get('url')
        if not verifyUrl(url):
            logging.info(f'BrowserActions::initialise: last page url was none from router, response=[{hist}]')
            url= self.startUrl
        logging.debug(f'going to: {url}')
        self.driver.get(url)
        ret = BrowserActions.Return(action=None, data=None, current_url=self.driver.current_url, name=self.name)
        caller.initialiseCallback(ret)


class BrowserService:
    retry_wait = 10
    retry_attempts = 10

    browser_action_cli_args = argparse.ArgumentParser()
    browser_process_command_queue = Queue()

    def __init__(self, attempts=0, *args, **kwargs):
        """
        Request a port of the nanny service and then start a webdriver session
        :param attempts: will recursively try to get a container, do not populate
        """
        if os.getenv('START_BROWSER', False):
            self.browser_monitor_thread = threading.Thread(target=BrowserService.beginBrowserThread, args=(self,))
            self.browser_monitor_thread.daemon = True
            self.browser_monitor_thread.start()
            sleep(10)
            logging.info(f'{type(self).__name__}::__init__(): Succesfully started browser process')
        self.driver_url = ''
        self.port = browser_params['port']
        url = f'http://{browser_params["host"]}:{self.port}/wd/hub'
        logging.info(f'browser host is set, using {url}')
        self.driver_url = url
        logging.info(f'Starting remote webdriver with {self.driver_url}')
        self.startWebdriverSession()
        logging.info(f'success')

    def driverHealthCheck(self):
        """
        request the current url from the driver and handle a no active session exception
        """
        try:
            self.driver.current_url
        # TODO here we should be able to detect if
        # 1. Session has expired, invalid session id or similiar
        # 2. process has stopped , we will get a urllib exception
        # 3. there has been a page crash ' WebDriverException
        except Exception as ex:
            logging.warning(f'{type(self).__name__}::driverHealthCheck(): Remote Webdriver session is unhealthy: error=[{type(ex).__name__}], args=[{ex.args}]')
            return False
        return True

    def startWebdriverSession(self):
        options = Options()
        options.add_argument("--headless")
        logging.info(f'starting webdriver session with {self.driver_url}')
        self.driver = webdriver.Remote(command_executor=self.driver_url,
                                       desired_capabilities=DesiredCapabilities.CHROME,
                                       options=options)
        logging.info("started webdriver session")

    def renewWebCrawler(self):
        logging.info(f'renewing webcrawler')
        self.driver.quit()
        self.startWebdriverSession()

    def renewDriverSession(self):
        self.renewWebCrawler()

    ###########################
    # following methods are for 
    # managing the selenium-st-
    # andalone-chrom process.

    def _bind_queue_and_log(queue: Queue, log):
        """
        iterate over next queue item or None and an io stream.
        """
        # zip together the next item on the queue or None and the next log line
        for i in log:
            queueItem = next_item = None if queue.empty() else queue.get(block=False)
            sleep(0.01)
            yield i, queueItem

    def beginBrowserThread(self):
        """
        open a process, forward its logs and watch for commands in a queue.
        """
        logging.info(f'BrowserService::beginBrowserThread(): Starting web browser thread')
        with subprocess.Popen(os.getenv('SELENIUM_PROCESS_SCRIPT',  "/opt/bin/start-selenium-standalone.sh"), stdout=subprocess.PIPE, bufsize=1, universal_newlines=True) as process:
            for line, command in BrowserService._bind_queue_and_log(self.browser_process_command_queue, process.stdout):
                logging.info(f'BrowserService:: {line}')
                if command is None:
                    logging.debug(f'nothing in queue')
                    continue
                elif command == 'KILL':
                    logging.debug(f'have queue item')
                    logging.info(f'BrowserService::beginBrowserThread(): Killing process {process.id}')
                    process.kill()
                    logging.debug(f'monitor thread is {"alive" if self.browser_monitor_thread.is_alive() else "complete"}')
        logging.info(f'Browser process {process.pid} has been torn down.')

    def _browser_clean_up(self):
        """
        place a clean up command to the thread monitoring the selenium process
        """
        logging.info(f'{type(self).__name__}::_browser_clean_up: sending kill command to browser monitor thread')
        self.browser_process_command_queue.put(blocking=True, item='KILL')




def reportParameter(parameter_key=None):
    endpoint = "http://{host}:{port}/parametermanager/reportParameter/{}/{}/{}".format(
        os.getenv("NAME"),
        parameter_key,
        "leader",
        **nanny_params
    )
    r.get(endpoint)





import docker
from unittest import TestCase
import time
import os
from pymongo import MongoClient
from feed.settings import mongo_params


browser_port = os.getenv('BROWSER_PORT', '4444')

base_params=dict(detach=True,
                 remove=True,
                 network='test')


sample_chain = dict(actions=[{"actionType": "CaptureAction", "css": ".card__body", "xpath": "//*[contains(concat( \" \", @class, \" \" ), concat( \" \", \"card__body\", \" \" ))]", "text": ""}], name='test_chain', startUrl='https://example.com', isRepeating=False)


class SeleniumTestInterface(TestCase):

    @classmethod
    def createSelenium(cls):
        cls.__client = docker.from_env()
        cls.__client.images.pull(repository='selenium/standalone-chrome', tag="3.141.59")
        # TODO should handle creation here
        cls.__selenium = cls.__client.containers.run(image='selenium/standalone-chrome:3.141.59', name='test_selenium',
                                                     ports={'4444/tcp': 4444}, **base_params)

    def killSelenium():
        __selenium = docker.from_env().containers.get('test_selenium')
        __selenium.kill()


class KafkaTestInterface(TestCase):

    @classmethod
    def createKafka(cls):
        cls.__client = docker.from_env()
        kafka_env = [
            "KAFKA_ZOOKEEPER_CONNECT=test_zookeeper:2181",
            "KAFKA_CREATE_TOPICS=\"sample-queue;worker-queue;worker-route;summarizer-route;leader-route\"",
            "KAFKA_OFFSETS_TOPIC_REPLICATION_FACTOR=1",
            "KAFKA_ADVERTISED_LISTENERS=PLAINTEXT://localhost:29092",
            "ALLOW_PLAINTEXT_LISTENER=yes"
        ]
        zookeeper_env = [
            "ZOOKEEPER_CLIENT_PORT=2181",
            "ALLOW_ANONYMOUS_LOGIN=yes",
            "ZOOKEEPER_TICK_TIME=2000"
        ]
        try:

            cls.__zookeeper = cls.__client.containers.run(image='confluentinc/cp-zookeeper',
                                                          name='test_zookeeper',
                                                          ports={'2181/tcp': 2181},
                                                          environment=zookeeper_env,
                                                          **base_params)
        except docker.errors.APIError as ex:
            if ex.status_code != 409:
                raise ex
        try:
            cls.__kafka = cls.__client.containers.run(image='confluentinc/cp-kafka',
                                                      name='test_kafka',
                                                      ports={'29092/tcp':29092},
                                                      environment=kafka_env,
                                                      **base_params)
            print('created kafka and zookeeper container')
        except docker.errors.APIError as ex:
            if ex.status_code != 409:
                raise ex

    def killKafka():
        __kafka = docker.from_env().containers.get('test_kafka')
        __zookeeper = docker.from_env().containers.get('test_zookeeper')
        __kafka.kill()
        __zookeeper.kill()


class MongoTestInterface:

    @classmethod
    def createMongo(cls):
        cls.__client = docker.from_env()
        # TODO should handle creation here
        print('Creating mongo image')
        try:
            cls.__mongo = cls.__client.containers.run(name='test_mongo',
                                                      image=os.getenv('MONGO_IMAGE','mongo'),
                                                      ports={'27017/tcp': 27017},
                                                      detach=True,
                                                      remove=True,
                                                      environment=[f'MONGO_INITDB_ROOT_USERNAME={os.environ["MONGO_USER"]}', f'MONGO_INITDB_ROOT_PASSWORD={os.environ["MONGO_PASS"]}'])
            time.sleep(4)
        except docker.errors.APIError as ex:
            if ex.status_code != 409:
                raise ex
        print('created docker done!')

        cls.mongo_client = MongoClient(**mongo_params)
        cls.mongo_client['actionChains']['actionChainDefinitions'].replace_one({'name': sample_chain.get('name')}, replacement=sample_chain, upsert=True)

    def killMongo():
        __container = docker.from_env().containers.get('test_mongo')
        __container.kill()


class PostgresTestInterface(TestCase):

    @classmethod
    def createPostgres(cls):
        cls.__client = docker.from_env()
        cls.__client.images.pull(repository='postgres', tag="latest")
        try:
            cls.__postgres = cls.__client.containers.run(name='test_database',
                                                          image='postgres',
                                                          ports={'5432/tcp': 5432},
                                                          environment=[f'POSTGRES_PASSWORD={os.environ["DATABASE_USER"]}', f'POSTGRES_USER={os.environ["DATABASE_PASS"]}'], **base_params)
        except docker.errors.APIError as ex:
            if ex.status_code != 409:
                raise ex

    def killPostgres():
        __postgres = docker.from_env().containers.get('test_database')
        __postgres.kill()

def ServiceFactory(component):
    """
    Create a service interface from a component name for test dependencies

    """

    def kill():
        __client = docker.from_env()
        try:
            __client.containers.get(f'test_{component}')
        except docker.errors.NotFound as ex:
            print(ex.args)
            pass


    def create():
        test_ports = {
            "ui-server": 5004,
            "nanny": 5003,
            "routing": 5002
        }
        environment = [
            "DATABASE_HOST=test_database",
            "DATABASE_PORT=5432",
            f'DATABASE_USER={os.getenv("DATABASE_USER")}',
            f'MONGO_HOST=test_mongo:27017',
            f'KAFKA_ADDRESS=localhost:29092',
            f'MONGO_PASS={os.getenv("MONGO_PASS")}'
            f'MONGO_USER={os.getenv("MONGO_USER")}'
            'ROUTER_HOST=test_router',
            'NANNY_HOST=test_nanny',
            'FLASK_PORT=5000'
        ]

        versions = {}
        with open(f'{os.getenv("DEPLOYMENT_ROOT")}/etc/manifest.txt', 'r') as manifest:
            for item in filter(lambda line: line.strip() != '',  manifest.read().split('\n')):
                versions.update({item.split("=")[0]: item.split('=')[1]})
        serviceVersion = versions.get(component)
        __client = docker.from_env()
        # TODO should handle creation here
        try:
            __container = __client.containers.run(f'{os.getenv("IMAGE_REPOSITORY")}/feed/{component}:{serviceVersion}', name=f'test_{component}', ports={f'{test_ports.get(component)}/tcp': 5000}, detach=True, remove=True,
                    environment=environment)
        except docker.errors.APIError as ex:
            if (ex.status_code != 409) and  (ex.status_code != 500):
                raise ex
    class TestInterface:
        pass

    setattr(TestInterface, f'create{component.capitalize()}', create)
    setattr(TestInterface, f'kill{component.capitalize()}', kill)

    return TestInterface





import docker
import inspect
import json
import logging
import re
from inspect import getfullargspec, getsource
from json import JSONDecodeError
from typing import List, Dict
from flask import Flask, Response, request, Request
import unittest
from unittest import TestCase, TestSuite
import requests as r

NOT_IMPLEMENTED = 509
WRONG_NUMBER_ARGS = 510
FAIL_CODE = 511
IGNORE_FAILS = True
TEST_CLAUSES = ["request", "response", "payload"]
MANDATORY = ["request", "response"]
JSON_CHARS = '}{:,"]['
METHOD_POS = 2  # if leading slash in doc example, method name will be in second chunk of request


class MethodNotImplemented(Response):
    def __init__(self):
        super().__init__('MethodNotImplemented', status=404)


class URLArgumentError(Response, Exception):
    def __init__(self, expected, actual):
        super().__init__(f'wrong number of args, expected: {expected}, actual: {actual}', status=404)


class FailResponse(Response, Exception):
    def __init__(self, reason, status = None):
        Response().__init__(reason, status = status if status is None else FAIL_CODE)



class ExpectedRequest:
    methods: List[str]
    cases: Dict[str, dict] = {}

    def __init__(self, methods: List[str] = None, cases=None, isMock=True):
        self.cases = cases
        self.methods = methods
        self.isMock = isMock
        self.actualResponse = None
        self.actualPayload = None
        self.actualCase = None
        self.trailing_slash = True

        for i in self.cases:
            case = self.cases.get(i)
            for clause in case:
                logging.info(f'_{clause} for method cases: {self.cases.keys()} has been implemented')
                if case is not None and all(ch in case.get(clause) for ch in JSON_CHARS):
                    try:
                        json.loads(case.get(clause))
                    except JSONDecodeError as e:
                        assert (False and f'{clause} of {i.split("/")[1]} for test case {i} is not valid. JSONDecodeError \
                                postion: {e.pos} line: {e.lineno} col: {e.colno} message: {e.msg}' or IGNORE_FAILS)

    @staticmethod
    def getmethodfrompath(uri):
        method = uri.split('/')[METHOD_POS]
        return method

    def getcasefrompath(self, uri):
        parts = list(filter(lambda item: len(item) > 0, uri.split("/")))[1:]
        method = f'/{"/".join(parts)}'
        if self.trailing_slash and len(parts) > 2:
            method = method + '/'
        return method

    def handleRequest(self):
        global request
        self.actualPayload = request.json() if 'json' in request.mimetype else None
        self.actualCase = self.getcasefrompath(request.path)
        self.actualMethod = request.method

        assert (request.method in self.methods and f'{self.actualCase} had invalid method usage, \
               actual was: {request.method}, expected: {self.methods}' or IGNORE_FAILS)
        try:
            self._request()
            self._payload()
        except (FailResponse, URLArgumentError) as e:
            return e

        return self._response()

    def _request(self):
        action = self.cases.get(self.actualCase)
        if action is None:
            return FailResponse(
                f'specific scenario not found actual: {self.actualCase}, expected one of: {self.cases.keys()}')

    def _response(self, testCase: TestCase, actual=None):
        # TODO actual is for the a test runner for testing the class given the documentation and not mocking
        caseDetails = self.cases.get(self.actualCase)
        if caseDetails is None:
            logging.error(False and f'{self.actualCase} is not implemented')
            return FailResponse('case not found')

        if self.actualMethod not in self.methods:
            return FailResponse('wrong method')
        mimetype = 'text' if any(c in self.cases.get(self.actualCase)['response'] for c in JSON_CHARS) else 'json'
        res = Response(self.cases.get(self.actualCase)['response'], status=200, mimetype=f'application/{type}')
        if not self.isMock:
            testCase.assertEqual(mimetype, res.mimetype, msg=f'wrong mimetype returned \
                    \n  expected: {type}\nactual: {res.mimetype}')
            testCase.assertEqual(res.json(), self.actualResponse.json(), msg=f'repsonse was invalid. \
                    \n  expected: {json.dumps(res.json(), indent=4)}\nactual: {json.dumps(self.actualResponse.json(), indent=4)}')
            return
        return res

    def _payload(self):
        try:
            case = self.cases[self.actualCase]
        except:
            logging.warning(f'{self.actualCase} not found')
            assert (f'{self.actualCase} not found' and False or IGNORE_FAILS)
            raise FailResponse(f'{self.actualCase} not found')
        if case.get('payload') is None:
            return

        payload = case.get('payload')
        logging.info(f'received payload to endpoint {self.actualCase}: {json.dumps(self.actualPayload)}')
        assert (self.actualPayload.keys() == payload.keys() and f'body of request {self.actualCase} was malformed. \
                \nactual: {json.dumps(request.json())}\nexpected: {json.dumps(payload)}' or IGNORE_FAILS)


class TestClass(TestCase):
    def __init__(self, actual, expected):
        super(TestCase, self).__init__()
        self.actual = actual
        self.expected = expected

    def test_json(self):
        self.assertListEqual(self.actual.keys(), self.expected.keys())

    def test_text(self):
        TestCase.assertEquals(self, self.actual, self.expected)

class DocumentationTest:
    _itemsregex = re.compile(f'(#{":|#".join(TEST_CLAUSES)}:)')
    _nameRegex = re.compile('<.*/>')
    _methodsRegex = re.compile('methods=\[')

    def __init__(self, func: callable):
        self.name = func
        self.source = str(func.__doc__)

    def run(self, self2):
        for i in self.getNames():
            clauses = self.getTestClauses(i)
            req = clauses.get('request').split('/')[1]
            res = self.name(self2, *req)
            if isinstance(res, Response):
                if res.mimetype == 'application/json':
                    test = TestClass(res.json(), clauses.get('response'))
                    test.run()

    @staticmethod
    def _testCaseRegex(name):
        return re.compile(f'(<{name}>|<{name}/>)')

    def getValidMethods(self):
        dec = inspect.getsource(self.name).split("\n")[0]
        default = ['GET']
        substrs = dec.split('methods')
        if len(substrs) == 1 or '@route' not in dec:  # if len of split is 1, methods not defined, use default
            return default
        substrs = substrs[1].split(']')
        i, loc = 0, None
        while i < len(substrs):
            if any(met in substrs[i] for met in ['GET', 'POST', 'DELETE', 'PUT']):
                loc = i
                break
            i += 1
        if loc is None:
            return default
        methods = substrs[loc].split(']')[0].replace('=', '').replace('[', '')
        methodList = methods.replace('"' if '"' in methods else "'", '').split(',')
        methodList = [met.strip() for met in methodList]
        return methodList

    def getTestClauses(self, name):
        res = DocumentationTest._testCaseRegex(name).split(self.source.replace("\n", " "))
        case = list(filter(lambda item: all(clause in item for clause in MANDATORY)
                                        and not all(ch in item for ch in '</>'), res))
        logging.info(f'found test case name: {name}')
        out, i = {}, 0
        clauses = case[0].split()
        while i < len(clauses):
            if any(clause in clauses[i] for clause in TEST_CLAUSES):
                # we found a test clause, following it is the details
                clause = clauses[i]
                det = ''
                i += 1 if len(clauses) > i else 0
                while not any(clau in clauses[i] for clau in TEST_CLAUSES) and i < len(clauses):
                    # we have split on spaces here so a dictionary will be fragmented. This while loop puts it
                    # back together
                    det += clauses[i]
                    i += 1
                    if i >= len(clauses):
                        break
                out.update({clause.replace('#', '').replace(':', ''): det.strip()})
                if i >= len(clauses):
                    break
            else:
                i += 1
                continue
        return out

    def getNames(self):
        res: List[str] = self._nameRegex.findall(str(self.source))
        return [r.strip("/<>") for r in res]

    @staticmethod
    def generate(func: callable) -> ExpectedRequest:
        # TODO when func has no arguments, we only can store one case as we store in map with the url as key.
        # TODO should have some sort of multimap or list of dicts
        doc = DocumentationTest(func)
        names = doc.getNames()
        caseResponses = {}
        for name in names:
            items = doc.getTestClauses(name)
            if len(items) != 0 and name is not None:
                for i in MANDATORY:
                    if items.get(i) is None:
                        assert (False and f'missing mandatory item {i} in testcase {name} for {func}')
                case = {}
                for i in TEST_CLAUSES:
                    if i == 'request':
                        req = items.get(i)
                        continue
                    if items.get(i) is None:
                        # implies non default clause
                        # TODO check that i is not in MANDATORY
                        continue
                    else:
                        case.update({i: items.get(i)})
                uri = req
                caseResponses.update({uri: case})
        methods = doc.getValidMethods()
        return ExpectedRequest(cases=caseResponses, methods=methods)


class MockedMethod:
    expectedRequest: ExpectedRequest

    def __init__(self, method):
        self.argspec = getfullargspec(method)
        self.method = method

    # TODO this method fails when running as expectedRequest is not defined when class is being initialised. should fix
    # def __str__(self):
    #   return f'{self.__class__.__name__}: {self.expectedRequest.methods}, \
    #           cases: {json.dumps(self.expectedRequest.cases, indent=4)}'

    def runMethod(self, req: Request, trailng_slash):
        self.expectedRequest = DocumentationTest.generate(self.method)
        self.expectedRequest.trailing_slash = trailng_slash
        return self.expectedRequest.handleRequest()

def getPublicMethods(service) -> List[str]:
    return [method for method in filter(lambda method: callable(getattr(service, method)) and '_' not in method,
                                     dir(service))]


def MockFactory(service, app):
    class MockService(service):

        actions: Dict[str, MockedMethod]

        def __init__(self, app):
            self.app = app
            self.cases = {}
            self.actions = {}

        def init(self):
            serviceMethods = [method for method in
                              filter(lambda method: callable(getattr(service, method)) and '_' not in method,
                                     dir(service))]
            for method in serviceMethods:
                args = getfullargspec(getattr(service, method))
                params = "/".join([f'<{pname}>' for pname in args[0][1:]])
                self._mockMethod(method)
                self.app.add_url_rule(f'/{service.get_route_base()}/{method}/{params}', method, self._runMethod)
            print(self.app.url_map)
        def _mockMethod(self, methodName: str):
            method = MockedMethod(getattr(service, methodName))
            self.actions.update({methodName: method})
            logging.debug(f'mocked method on uri: {methodName} with {method}')

        def _runMethod(self, **args):
            uri = request.path
            action = self.actions.get(ExpectedRequest.getmethodfrompath(uri))
            logging.debug(f'running mocked method on uri: {uri} with {action}')
            if action is None:
                return FailResponse('no action')
            compare = len(action.argspec.args) - 1 if 'self' in action.argspec.args else len(action.argspec.args)
            if compare != len(args):
                return URLArgumentError(actual=list(args.values()), expected=action.argspec)
            return action.runMethod(request, trailng_slash=service.trailing_slash)

    return MockService(app)

def makeTestCase(method):
    def test_case(self):
        expectedRequest = DocumentationTest.generate(getattr(service, method))
        for case in expectedRequest.cases:
            method = expectedRequest.methods[0]
            req = r.request(method, f'http://localhost:{self.runningOn}/{case}',
                            json=expectedRequest.cases.get(case).get('payload'))
            expectedRequest._response(self, req)
    return test_case

# TODO check 25.3.7.3.1. load_tests Protocol on python docs
def TestFactory(service, runningOn: str):

    class Case(TestCase):
        methods = getPublicMethods(service)
        def __init__(self):
            self.runningOn = runningOn
            for method in self.methods:
                print( f'test_{method}')
                setattr(self, f'test_{method}', makeTestCase(method))
            super().__init__()
    testSuite
    suite = TestSuite()
    suite.addTest(suite)

def init_mongo():
    client = docker.from_env()
    client.containers.run("config_data", network=os.getenv("NETWORK", "feed_default"))

if __name__ == "__main__":
    from ui_server.src.main.feedmanager import FeedManager
    testSuite = TestFactory(FeedManager, 5000)
    testSuite.test_getParameter()




import os

# service params

kafka_params = {
    "bootstrap_servers": [os.getenv("KAFKA_ADDRESS", "localhost:29092")],
}

browser_params = {
    "internal_port":  4444,
    "host": os.getenv("BROWSER_CONTAINER_HOST", 'localhost'),
    "image": os.getenv('BROWSER_IMAGE', 'selenium/standalone-chrome:3.141.59'),
    "port": os.getenv('BROWSER_PORT', 4444),
    "max": os.getenv("MAX_FEEDS", 10),
    "base_port": os.getenv("BROWSER_BASE_PORT", 4444)
}


mongo_params = {
    "host": os.getenv("MONGO_HOST", "localhost:27017"),
    "username": os.getenv("MONGO_USER", "root"),
    "password": os.getenv("MONGO_PASS", "root"),
    "serverSelectionTimeoutMS": 5
}

database_parameters = {
    "host": os.getenv("DATABASE_HOST", "localhost"),
    "port": os.getenv("DATABASE_PORT", 5432),
    "database": os.getenv("DATABASE_NAME", "feeds"),
    "user": os.getenv("DATABASE_USER", "feeds"),
    "password": os.getenv("DATABASE_PASS", "postgres"),
}


########################
# COMPONENT Connectivity
# defaults to localhost for local debugging
# ports here should match up with the development.yml for exposing ports

ui_server_params = {
    "host": os.getenv('UISERVER_HOST', 'localhost'),
    "port": os.getenv('FLASK_PORT', 5004) # ui-server
    }
nanny_params = {
    "host": os.getenv("NANNY_HOST", "localhost"),
    "port": os.getenv("FLASK_PORT", 5003), # nanny
    "api_prefix": "containercontroller",
    "params_manager": "parametercontroller"
}

routing_params = {
    "host": os.getenv("ROUTER_HOST", "localhost"),
    "port": os.getenv("FLASK_PORT", 5002), # routing
    "api_prefix": "routingcontroller"
}

persistence_params = {
    "host": os.getenv("PERST_HOST", "localhost"),
    "port": os.getenv("FLASK_PORT", 5006) # persistence
}

summarizer_params = {
    "host": os.getenv("SUMMARIZER_HOST", "localhost"),
    "port": os.getenv("FLASK_PORT", 5005) # summarizer
}

command_params = {
    "host": os.getenv("COMMANDS_HOST", "localhost"),
    "port": os.getenv("FLASK_PORT", 5001) # commands
}
########################


# settings

retry_params = {
    "times": 10,
    "wait": 10
}

feed_params = {
    "image": os.getenv("LEADER_TEMPLATE"),
    "success": os.getenv("LEADER_START", "feed has started"),
    "base_port": int(os.getenv("LEADER_BASE_PORT", 9000))
}


class BrowserConstants:
    CONTAINER_TIMEOUT = int(os.getenv('CONTAINER_TIMEOUT', 10))
    CONTAINER_SUCCESS = 'Selenium Server is up and running on port'
    CONTAINER_QUIT = "Shutdown complete"
    client_connect = 'wd/hub'
    worker_timeout = int(os.getenv('WORKER_TIMEOUT', '3'))



from unittest import TestCase
import logging
import unittest

from feed.crawling import BrowserService, BrowserActions
from feed.actionchains import CaptureAction
from feed.actionchains import CaptureAction, ClickAction, PublishAction, InputAction
from feed.testinterfaces import SeleniumTestInterface

#from feed.testutils import MockFactory # in order to mock the routing factory we will need the application code
# other option: build a test image which runs the mock?
# an environment variable and an image which pulls the application code to a certain version and mocks it?

test_action_chain = {}


crawling = logging.getLogger('feed.crawling')
actions = logging.getLogger('feed.actionchains')
sh = logging.FileHandler('/home/rory/app/feed/tmp/logs/feed.crawling.logs')
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s -%(message)s')
sh.setLevel(logging.DEBUG)
sh.setFormatter(formatter)
crawling.addHandler(sh)
actions.addHandler(sh)


class TestBrowserActions(SeleniumTestInterface):

    @classmethod
    def setUpClass(cls):
        SeleniumTestInterface.create()

    def setUp(cls):
        cls.browserService = BrowserService()
        cls.actionChain = BrowserActions(driver=cls.browserService.driver, **test_action_chain)

    def tearDown(cls):
        cls.browserService.driver.quit()
        del cls.browserService

    def test_onCaptureAction(cls):
        action = CaptureAction(position=0, captureName='donedeal_test', **{"actionType": "CaptureAction", "css": ".card__body", "xpath": "//*[contains(concat( \" \", @class, \" \" ), concat( \" \", \"card__body\", \" \" ))]", "text": ""})
        cls.actionChain.driver.get("https://www.donedeal.co.uk/cars")
        ret = cls.actionChain.onCaptureAction( action)
        for item in ret:
            print(item.data)
            cls.assertIsNotNone(item)
        cls.failIf(len(ret) is 0)

    def test_onPublishAction(cls):
        action = PublishAction(position=0, **{"actionType": "PublishAction", "css": ".card__body", "xpath": "//*[contains(concat( \" \", @class, \" \" ), concat( \" \", \"card__body\", \" \" ))]", 'text': '', 'isSingle': False, 'urlStub': 'https://www.donedeal.ie/cars-for-sale'})
        cls.actionChain.driver.get("https://www.donedeal.co.uk/cars")
        ret = cls.actionChain.onPublishAction(action)
        for item in ret:
            cls.failIf('https://' not in ret.data)
        cls.failIf(len(ret) == 0)

    #TODO on input action tests

    def test_onClickAction(cls):
        startUrl = "https://www.donedeal.co.uk/cars"
        cls.actionChain.driver.get(startUrl)
        action = ClickAction(position=0, **{"actionType": "ClickAction", "css": ".ng-isolate-scope", "text": "Next", 'xpath':'//*[contains(concat( " ", @class, " " ), concat( " ", "ng-isolate-scope", " " ))]'}) # assumes were clicking on only one thing for the time being
        cls.actionChain.onClickAction(action)
        cls.assertNotEqual(cls.actionChain.driver.current_url, startUrl)






import sys
import time
import json
import requests as r
import logging
from flask_classy import FlaskView



class Service(FlaskView):

    def healthCheck(self):
        return "ok"


class Client:
    max_attempts = 10
    wait = 10

    def __init__(self, name, attempts=0, **params):
        try:
            hc = r.get("http://{host}:{port}/service/healthCheck".format(**params))
            if 100 < hc.status_code < 300:
                self.health = True
                logging.info(f'connected to {name}')
            if hc.status_code is 404:
                logging.warning(f'health check for {name}/{params.get("api_prefix")} was not found')
                raise Exception()
            else:
                logging.info("")
                self.health = False
        except Exception as e:
            attempts += 1
            if attempts < self.max_attempts:
                time.sleep(self.wait)
                logging.warning(f'could not connect to {name}, trying again {self.max_attempts - attempts} more times')
                self.__init__(name, attempts, **params)
            else:
                logging.error(f'could not connect to {name}. Giving up. \nParameters were:  \n{json.dumps(params, indent=4)}')
                sys.exit()




from unittest import TestCase
from flask_classy import route

from feed.testutils import DocumentationTest

example1 = {
    "request": "/getFeeds/",
    "response": '["donedeal","pistonheads"]'
}

example2 = {
    "request": "/getFeeds/",
    "payload": '{"key":"value"}',
    "response": '["donedeal","pistonheads"]'
}


@route(rule="/a/partial/url/<params:string>", methods=['GET', 'POST'], other='somethingelse')
def getFeeds(string):
    """
    <example1>
        #request: /getFeeds/case1
        #response: ["donedeal", "pistonheads"]
    <example1/>

    <example2>
        #request: /getFeeds/case2
        #payload: {"key": "value"}
        #response: ["donedeal", "pistonheads"]
    <example2/>

    :return:
    """
    pass


class TestDocumentationTest(TestCase):
    def test__test_case_regex(self):
        regex = DocumentationTest._testCaseRegex('exampleName')
        # self.assert_(isinstance(regex,  _sre.SRE_Pattern))
        self.assertTrue(regex.pattern == '(<exampleName>|<exampleName/>)')

    def test_getValidMethods(self):
        docTest = DocumentationTest(getFeeds)
        mets = docTest.getValidMethods()
        self.assertListEqual(mets, ['GET', 'POST'], msg='list of methods do not match')

    def test_get_test_clauses(self):
        docTest = DocumentationTest(getFeeds)

        # request and response can be parsed
        mets = docTest.getTestClauses('example1')
        self.assertDictEqual(mets, example1)

        # optionally payload may be parsed
        # TODO parse json in this method
        mets = docTest.getTestClauses('example2')
        self.assertDictEqual(mets, example2)

    def test_get_names(self):
        docTest = DocumentationTest(getFeeds)
        names = docTest.getNames()
        self.assertListEqual(['example1', 'example2'], names)

    def test_generate(self):
        er = DocumentationTest.generate(getFeeds)
        self.assertDictEqual(er.cases, {'/getFeeds/case1': {
                                            'response': '["donedeal","pistonheads"]'
                                        },
                                        '/getFeeds/case2': {
                                            'response': '["donedeal","pistonheads"]',
                                            'payload': '{"key":"value"}'}
                                        })
        # TODO case where there are multiple cases for a method with no params
